#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ddnsd-conf.h"

#define LINE_SIZE 1024

typedef struct BcmDDnsListItems { char **vals;
                                  struct BcmDDnsListItems *next; } BcmDDnsListItems;

typedef BcmDDnsListItems * pBcmDDnsListItems;

typedef struct BcmDDnsList { int num_vals;
                             pBcmDDnsListItems items; } BcmDDnsList;

typedef BcmDDnsList * pBcmDDnsList;

static pBcmDDnsList ddnsConfig = NULL;
static pBcmDDnsList ddnsCache = NULL;

static int BcmDDnsList_add( pBcmDDnsList, char **vals );
static void BcmDDnsList_close( pBcmDDnsList );

int
BcmDDnsConfig_init( char *pszFilename ) {
   FILE *fp;
   int val_idx;
   char line[LINE_SIZE], *value, *variable;
   int line_num = 0;
   char *vals[5];
   char *val_names[] = { "interface", "service", "hostname", "username", "password", NULL };

   if( ( fp = fopen( pszFilename, "r" ) ) == NULL ) {
     return -1;
   }

   if ( ( ddnsConfig = malloc( sizeof( BcmDDnsList ) ) ) == NULL ) {
     fclose( fp );
     return -1;
   }
   printf("DDNSConfig_init %s\n", pszFilename);
   ddnsConfig->num_vals = 5;

   memset( vals, 0, sizeof( vals ) );

   while( fgets( line, LINE_SIZE, fp ) != NULL ) {
     line_num++;
     switch( line[0] ) {
     case '\n':
     case '#':
     case ';':
       continue;  // Skip comments
     case '[': // New interface
       // Verify the last interface had all the values
       if( !vals[IFACE] || !vals[HOSTNAME] || !vals[SERVICE] || !vals[USERNAME] || !vals[PASSWORD] ) {
         // fprintf( stderr, "Discarding interface with missing information: Missing " );

         for( val_idx = 0; val_idx < 5; val_idx++ ) {
           if( vals[val_idx] )
             free( vals[val_idx] );
         }
       } else {
         BcmDDnsList_add( ddnsConfig, vals );
//         memset( vals, 0, sizeof( vals ) );
       }

       memset( vals, 0, sizeof( vals ) );

       value = strtok( line+1, "]\n" );

       if( !strlen( value ) )
         continue;

       vals[HOSTNAME] = strdup( value );
       break;
     default:
       variable = strtok( line, "=\n" );
       value = strtok( NULL, "\n" );
       if ( strlen( value ) ) {
         if ( !strcmp( variable, "interface" ) ) {
           if( vals[IFACE] )
             free( vals[IFACE] );
           vals[IFACE] = strdup( value );
         } else if( !strcmp( variable, "username" ) ) {
           if( vals[USERNAME] )
             free( vals[USERNAME] );
           vals[USERNAME] = strdup( value );
         } else if ( !strcmp( variable, "password" ) ) {
           if( vals[PASSWORD] )
             free( vals[PASSWORD] );
           vals[PASSWORD] = strdup( value );
         } else if ( !strcmp( variable, "service" ) ) {
           if( !strcmp( value, "tzo" ) && !strcmp( value, "dyndns" ) &&
               !strcmp( value, "dtdns" ) && !strcmp( value, "noip" ) ) {
             fprintf( stderr, "Service %s not supported. (Line %d)\n", value, line_num );
             continue;
           }
           if( vals[SERVICE] )
             free( vals[SERVICE] );
           vals[SERVICE]= strdup( value );
         } else {
           fprintf( stderr, "Invalid key %s on line %d.\n", variable, line_num );
         }
       } 
     }
   }

   // P.T. clean up resource
   fclose( fp );

   if( !vals[IFACE] || !vals[HOSTNAME] || !vals[SERVICE] || !vals[USERNAME] || !vals[PASSWORD] ) {
     // fprintf( stderr, "Discarding interface with missing information: Missing " );

     for( val_idx = 0; val_idx < 5; val_idx++ ) {
       if( vals[val_idx] )
         free( vals[val_idx] );
     }

   } else {
     BcmDDnsList_add( ddnsConfig, vals );
   }
}

static int
BcmDDnsList_add( pBcmDDnsList ddnsList, char **vals ) {
  pBcmDDnsListItems to_add, list;

  //printf("DDNSList_add \n");
  if( ( to_add = malloc( sizeof( BcmDDnsListItems ) ) ) == NULL ) {
    return -1;
  }
  if( ( to_add->vals = malloc( sizeof( char * ) * ddnsList->num_vals ) ) == NULL ) {
    free( to_add );
    return -1;
  }

  to_add->next = NULL;
  memcpy( to_add->vals, vals, ddnsList->num_vals * sizeof( char * ) );
  if( ddnsList->items == NULL ) {
    ddnsList->items = to_add;
  } else {
    for( list = ddnsList->items; list->next != NULL; list = list->next ) ;
    list->next = to_add;
  }
  return 0;
}

void
BcmDDnsConfig_iter( void (*func)(char *) ) {
  pBcmDDnsListItems list;

  //printf("DDNSCache_iter\n");
  for( list = ddnsConfig->items; list != NULL; list = list->next ) {
    func( list->vals[HOSTNAME] );
  }
}

void
BcmDDnsConfig_remove( char *hostname ) {
  pBcmDDnsListItems list;
  int i;
  pBcmDDnsListItems prev = NULL;
  
  //printf("DDNSConfig_remove %s\n", hostname);                                                                              
  for (list = ddnsConfig->items; list != NULL; prev = list, list = list->next ) {
    if ( !strcmp( hostname, list->vals[HOSTNAME] ) ) {
      for( i = 0; i < ddnsConfig->num_vals; i++ ) {
        free( list->vals[i] );
      }
      free( list->vals );
      if (prev) 
        prev->next = list->next;
      else
        ddnsConfig->items = list->next;
      free( list );
      return;
    }
  } 
}

char *
BcmDDnsConfig_getInterface( char *hostname ) {
  pBcmDDnsListItems list;
  int i;
  char **interfaces;

  //printf("DDNSCOnfig_getinterface %s\n", hostname);
  for (list = ddnsConfig->items; list != NULL; list = list->next ) {
    if ( !strcmp( hostname, list->vals[HOSTNAME] ) ) {
      return list->vals[IFACE];
    }
  }

#if 0
  for (i = 0, list = ddnsConfig->items; list != NULL; list = list->next ) {
    if ( !strcmp( hostname, list->vals[HOSTNAME] ) ) {
      i++;
    }
  }
  if( !i ) {
    return NULL;
  } else {
    if( ( interfaces = malloc( (i+1) * sizeof( char * ) ) ) == NULL ) {
      return NULL;
    } else {
      for (i = 0, list = ddnsConfig->items; list != NULL; list = list->next ) {
        if ( !strcmp( hostname, list->vals[HOSTNAME] ) ) {
          interfaces[i] = list->vals[IFACE];
          i++;
        }
      }
    }
  }
  return interfaces;
#endif
}

char *
BcmDDnsConfig_getService( char *hostname ) {
  pBcmDDnsListItems list;
  int i;
  char **interfaces;

  printf("DDNSConfig_getservice %s \n", hostname);
  for (list = ddnsConfig->items; list != NULL; list = list->next ) {
    if ( !strcmp( hostname, list->vals[HOSTNAME] ) ) {
      return list->vals[SERVICE];
    }
  }
}

char *
BcmDDnsConfig_getUsername( char *hostname ) {
  pBcmDDnsListItems list;
  int i;
  char **interfaces;

  printf("DDNSConfig_getUsername %s \n", hostname);
  for (list = ddnsConfig->items; list != NULL; list = list->next ) {
    if ( !strcmp( hostname, list->vals[HOSTNAME] ) ) {
      return list->vals[USERNAME];
    }
  }
}

char *
BcmDDnsConfig_getPassword( char *hostname ) {
  pBcmDDnsListItems list;
  int i;
  char **interfaces;

  printf("DDNSConfig_getPassword %s \n", hostname);
  for (list = ddnsConfig->items; list != NULL; list = list->next ) {
    if ( !strcmp( hostname, list->vals[HOSTNAME] ) ) {
      return list->vals[PASSWORD];
    }
  }
}


void
BcmDDnsConfig_close( void ) {
  printf("DDNSConfig_Close\n");
   BcmDDnsList_close( ddnsConfig );
   ddnsConfig = NULL;
}


void
BcmDDnsList_close( pBcmDDnsList ddnsList ) {
  pBcmDDnsListItems prev=NULL, list;
  int i;

  //printf("DDNSList_close\n");
  for( list = ddnsConfig->items; list != NULL; list = list->next ) {
    if (prev != NULL )
      free( prev );
    prev = list;
    for( i = 0; i < ddnsList->num_vals; i++ )
      if( list->vals[i] != NULL )
        free( list->vals[i] );
    free( list->vals );
  }
  if( prev != NULL )
    free( prev );
  ddnsList->items = NULL;
  free( ddnsList );
}

int
BcmDDnsCache_init( char *filename ) {
  char line[LINE_SIZE], *vals[2];
  FILE *fp;

  printf("DDNSCache_init %s \n", filename);
  if ( (ddnsCache = malloc( sizeof( pBcmDDnsList ) ) ) == NULL ) {
    return -1;
  }
  ddnsCache->num_vals = 2;
  ddnsCache->items = NULL;

  if ((fp = fopen( filename, "r" ) ) == NULL )
    return 0;

  while((fgets(line, LINE_SIZE, fp ) ) != NULL ) {
    vals[HOSTNAME] = strdup( strtok(line, "," ) );
    vals[CACHE_IP] = strtok( NULL, "\n" );
    //printf("cache_init: %s, %s\n", vals[HOSTNAME], vals[CACHE_IP]);
    BcmDDnsList_add( ddnsCache, vals );
  }

  // P.T. clean up resource
  fclose( fp );
  return 0;
}

char *
BcmDDnsCache_get( char *hostname ) {
  pBcmDDnsListItems list;

  //printf("DDNSCache_get %s \n", hostname);
  if (ddnsCache == NULL ) {
    return NULL;
  }
  
  for( list = ddnsCache->items; list != NULL; list = list->next ) {
    if ( !strcmp( hostname, list->vals[HOSTNAME] ) ) {
      return list->vals[CACHE_IP];
    }
  }
  return NULL;
}

void
BcmDDnsCache_update( char *hostname, char *address ) {
  pBcmDDnsListItems list;
  char *vals[2];

  if (ddnsCache == NULL )
    return;
  
  for( list = ddnsCache->items; list != NULL; list = list->next ) {
    if (!strcmp( hostname, list->vals[HOSTNAME] ) ) {
      if (list->vals[CACHE_IP] )
        free(list->vals[CACHE_IP]);
      list->vals[CACHE_IP] = strdup( address );
      return;
    }
  }
  vals[HOSTNAME] = strdup( hostname );
  vals[CACHE_IP] = strdup( address );
  printf("CACHE_UPDATE: %s, %s\n", vals[HOSTNAME], vals[CACHE_IP]);
  BcmDDnsList_add( ddnsCache, vals );
}

void
BcmDDnsCache_save( char *pszFilename ) {
  pBcmDDnsListItems list;
  FILE *fp;

  if (ddnsCache == NULL || ddnsCache->items == NULL ) {
    return;
  }

  if( ( fp = fopen( pszFilename, "w" ) ) == NULL ) {
    return;
  }

  for (list = ddnsCache->items; list != NULL; list = list->next ) {
    fprintf( fp, "%s,%s\n", list->vals[HOSTNAME], list->vals[CACHE_IP] );
    printf("ddnsCache_save: %s,%s\n", list->vals[HOSTNAME], list->vals[CACHE_IP]);
  }
  fclose( fp );
}

void
BcmDDnsCache_close() {
  //printf("DDNSCache_close\n");
   BcmDDnsList_close( ddnsCache );
   ddnsCache = NULL;
}

/*
#ifdef CONFIG_MAIN
int main(int argc, char **argv ) {
  pBcmDDnsListItems list;
  int i;

  printf("config_main!\n");
  BcmDDnsConfig_init( argv[1] );
  if (ddnsConfig == NULL || ddnsConfig->items == NULL) {
    printf("No interfaces defined.\n");
  } else {
    for( i = 0, list = ddnsConfig->items; list != NULL; list = list->next, i++ ) {
      printf( "Interface (%2d): %s\n", i, list->vals[IFACE] );
      printf( "Hostname      : %s\n", list->vals[HOSTNAME] );
      printf( "Service       : %s\n", list->vals[SERVICE] );
      printf( "Username      : %s\n", list->vals[USERNAME] );
      printf( "Password      : %s\n", list->vals[PASSWORD] );
      printf( "\n" );
    }
  }
}

#elif CACHE_MAIN
int main(int argc, char **argv ) {
  pBcmDDnsListItems list;

  printf("cache_main!\n");
  BcmDDnsCache_init( argv[1] );  
  if (ddnsCache == NULL ) {
    puts( "Problem." );
    exit(1);
  }
  if( ddnsCache->items == NULL ) {
    puts("No interfaces defined.");
  } else {
    for (list = ddnsCache->items; list != NULL; list = list->next ) {
      printf( "%s cached as %s.\n", list->vals[HOSTNAME], list->vals[CACHE_IP] );
    }
  }
  BcmDDnsCache_update( "somewhere.not.who", "209.1.1.1" );
  for (list = ddnsCache->items; list != NULL; list = list->next ) {
    printf( "%s cached as %s.\n", list->vals[HOSTNAME], list->vals[CACHE_IP] );
  }

}
#endif*/
